#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/param.h>
#include "hxd.h"

#if defined(CONFIG_EXEC)

static void
exec_close (int fd)
{
	close(fd);
	memset(&hxd_files[fd], 0, sizeof(struct hxd_file));
	FD_CLR(fd, &hxd_rfds);
	if (high_fd == fd) {
		for (fd--; fd && !FD_ISSET(fd, &hxd_rfds); fd--)
			;
		high_fd = fd;
	}
}

static void
exec_ready_read (int fd)
{
	ssize_t r;
	u_int8_t buf[16384];
	u_int16_t uid;
	
	uid = htons(mangle_uid(hxd_files[fd].conn.htlc));
	r = read(fd, buf, sizeof(buf));
	if (r <= 0)
		exec_close(fd);
	else {
		/* changed the line below to send a private message to the user rather
		 * than output to chat. Sends the message with the same UID as the user
		 * processing the command so that if they hit the reply button, it will
		 * send the message to themselves (rather than to a non-existant user.
		 *  --Devin*/
		hlwrite(hxd_files[fd].conn.htlc, HTLS_HDR_MSG, 0, 3,
			HTLS_DATA_UID, sizeof(uid), &uid,
			HTLS_DATA_MSG, r, buf,
			HTLS_DATA_NAME, 3, "hxd");
	}
}

#if 0 /* add somethin like http post here?? */
static void
exec_ready_write (int fd)
{
}
#endif

static void
cmd_exec (struct htlc_conn *htlc, char *command)
{
	int argc, pfds[2];
	char *argv[32], *p, *pii, *thisarg, cmdpath[MAXPATHLEN];

	while ((p = strstr(command, "../"))) {
		for (pii = p; *pii; pii++)
			*pii = *(pii + 3);
	}

	for (argc = 0, thisarg = p = command; *p && argc < 30; p++) {
		if (*p == ' ') {
			*p = 0;
			argv[argc++] = thisarg;
			thisarg = p + 1;
		}
	}
	if (thisarg != p)
		argv[argc++] = thisarg;
	argv[argc] = 0;
	snprintf(cmdpath, sizeof(cmdpath), "%s/%s", hxd_cfg.paths.exec, argv[0]);

#if 0
	hxd_log("%s execs %s", htlc->name, cmdpath);
#endif
	if (pipe(pfds)) {
		hxd_log("cmd_exec: pipe: %s", strerror(errno));
		return;
	}
	if (pfds[0] >= hxd_open_max) {
		hxd_log("%s:%d: %d >= hxd_open_max (%d)", __FILE__, __LINE__, pfds[0], hxd_open_max);
		close(pfds[0]);
		close(pfds[1]);
		return;
	}
	switch (fork()) {
		case -1:
			hxd_log("cmd_exec: fork: %s", strerror(errno));
			close(pfds[0]);
			close(pfds[1]);
			return;
		case 0:
			if (dup2(pfds[1], 1) == -1 || dup2(pfds[1], 2) == -1) {
				hxd_log("cmd_exec: dup2: %s", strerror(errno));
				_exit(1);
			}
			close(0);
#if 0
			{
				int i;

				fprintf(stderr, "executing");
				for (i = 0; i < argc; i++)
					fprintf(stderr, " %s", argv[i]);
			}
#endif
			{
				char *envp[2], rootdir[MAXPATHLEN + 16];

				snprintf(rootdir, sizeof(rootdir), "ROOTDIR=%s", htlc->rootdir);
				envp[0] = rootdir;
				envp[1] = 0;
				execve(cmdpath, argv, envp);
			}
			fprintf(stderr, "\r%s: %s", argv[0], strerror(errno));
			_exit(1);
		default:
			close(pfds[1]);
			hxd_files[pfds[0]].conn.htlc = htlc;
			hxd_files[pfds[0]].ready_read = exec_ready_read;
			FD_SET(pfds[0], &hxd_rfds);
			if (high_fd < pfds[0])
				high_fd = pfds[0];
			break;
	}
}
#endif /* ndef CONFIG_EXEC */

static void
update_user (struct htlc_conn *htlc)
{
	struct htlc_conn *htlcp;
	u_int16_t uid, icon16, color;

	uid = htons(mangle_uid(htlc));
	icon16 = htons(htlc->icon);
	color = htons(htlc->color);
	for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
		if (!htlcp->access_extra.user_getlist)
			continue;
		hlwrite(htlcp, HTLS_HDR_USER_CHANGE, 0, 4,
			HTLS_DATA_UID, sizeof(uid), &uid,
			HTLS_DATA_ICON, sizeof(icon16), &icon16,
			HTLS_DATA_COLOUR, sizeof(color), &color,
			HTLS_DATA_NAME, strlen(htlc->name), htlc->name);
	}
}

void
toggle_away (struct htlc_conn *htlc)
{
	htlc->color = htlc->color & 1 ? htlc->color-1 : htlc->color+1;
	if (!htlc->flags.visible)
		return;
#if defined(CONFIG_SQL)
	sql_modify_user(htlc->name, htlc->icon, htlc->color, htlc->uid);
#endif
	update_user(htlc);
}

static void
cmd_broadcast (struct htlc_conn *htlc, char *chatbuf)
{
	u_int16_t style, len;
	struct htlc_conn *htlcp;

	if (!htlc->access.can_broadcast)
		return;

	style = htons(1);
	len = strlen(chatbuf);
	for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next)
		hlwrite(htlcp, HTLS_HDR_MSG, 0, 2,
			HTLS_DATA_STYLE, 2, &style,
			HTLS_DATA_MSG, len, chatbuf);
}

/* added the cmd_alrt method to send a broadcast to a
 * certain user. --Devin */

static void
cmd_alrt (struct htlc_conn *htlc, char *chatbuf)
{
	char *p, *str;
	u_int32_t uid;
	u_int16_t style, len;
	struct htlc_conn *htlcp;
	int n;
	
	if (!htlc->access.can_broadcast)
		return;

	p = chatbuf;
	uid = atou32(p);
	if (!uid)
		return;
	htlcp = isclient(htlc->sid, uid);
	if (!htlcp)
		return;
	n = 1;
	while (*p && *p != ' ') {
		p++;
		n++;
	}
	style = htons(1);
	str = &chatbuf[n];
	len = strlen(str);
	hlwrite(htlcp, HTLS_HDR_MSG, 0, 2,
		HTLS_DATA_STYLE, 2, &style,
		HTLS_DATA_MSG, len, str);
}

static void
cmd_visible (struct htlc_conn *htlc)
{
	u_int16_t uid16;
	struct htlc_conn *htlcp;

	if (!htlc->access_extra.user_visibility)
		return;

	htlc->flags.visible = !htlc->flags.visible;
	if (!htlc->flags.visible) {
		uid16 = htons(mangle_uid(htlc));
		for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
			if (!htlcp->access_extra.user_getlist)
				continue;
			hlwrite(htlcp, HTLS_HDR_USER_PART, 0, 1,
				HTLS_DATA_UID, sizeof(uid16), &uid16);
		}
	} else {
		update_user(htlc);
	}
}

static void
cmd_color (struct htlc_conn *htlc, char *chatbuf)
{
	u_int16_t color = htlc->color;

	if (!htlc->access_extra.user_color)
		return;

	htlc->color = strtoul(chatbuf, 0, 0);
#if defined(CONFIG_SQL)
	sql_modify_user(htlc->name, htlc->icon, htlc->color, htlc->uid);
#endif
	if (htlc->color != color && htlc->flags.visible)
		update_user(htlc);
}

extern void access_extra_set (struct extra_access_bits *acc, char *p, int val);

static void
cmd_access (struct htlc_conn *htlc, char *chatbuf)
{
	char *p, *str;
	u_int32_t uid;
	int val;
	struct htlc_conn *htlcp;

	if (!htlc->access_extra.user_access)
		return;

	p = chatbuf;
	uid = atou32(p);
	if (!uid)
		return;
	htlcp = isclient(htlc->sid, uid);
	if (!htlcp)
		return;
	if (!htlcp->access_extra.access_volatile)
		return;
	while (*p && *p != ' ')
		p++;
	while (*p && *p == ' ')
		p++;
	while (*p) {
		str = p;
		p = strchr(p, '=');
		if (!p)
			break;
		*p = 0;
		p++;
		val = *p == '1' ? 1 : 0;
		p++;
		access_extra_set(&htlcp->access_extra, str, val);
		while (*p && *p != ' ')
			p++;
		while (*p && *p == ' ')
			p++;
	}
}

/* 
 * added cmd_mkpuppet method to produce a fake user.
 * currently unfinished and not completely functional.
 * creates a user of UID 65535 that is added to the userlist.
 * note: the user cannot be 0wned, and cannot be dosconnected.
 * the user does not show to new users.
 * --Devin
*/

static void
cmd_mkpuppet (char *name)
{
	struct htlc_conn *htlcp;
	u_int16_t len, color, icon16;
	
	len = strlen(name);
	for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
		if (!htlcp->access_extra.user_getlist)
			continue;
		icon16 = htons(htlcp->icon);
		color = htons(htlcp->color);
        hlwrite(htlcp, HTLS_HDR_USER_CHANGE, 0, 4,
			HTLS_DATA_UID, 2, "\xFF\xFF",
			HTLS_DATA_ICON, sizeof(icon16), &icon16,
			HTLS_DATA_COLOUR, sizeof(color), &color,
			HTLS_DATA_NAME, len, name);
	}
}

static int
user_0wn (struct htlc_conn *htlc, char *str, char *p)
{
	if (!strcmp(str, "icon")) {
		u_int16_t icon;

		icon = strtoul(p, 0, 0);
		htlc->icon = icon;
		return 1;
	} else if (!strcmp(str, "color")) {
		u_int16_t color;

		color = strtoul(p, 0, 0);
		htlc->color = color;
		return 2;
	} else if (!strcmp(str, "name")) {
		size_t len;

		/* copies the rest of the command into name! */
		len = strlen(p);
		if (len > sizeof(htlc->name)-1)
			len = sizeof(htlc->name)-1;
		memcpy(htlc->name, p, len);
		htlc->name[len] = 0;
		return 3;
	} else if (!strcmp(str, "g0away") || !strcmp(str, "visible")) {
		/* added the above if to add the feature to make
		 * users g0away --Devin */
		
		int val;
		char *str;

		str = "user_visibility";
		val = htlc->access_extra.user_visibility;
		access_extra_set(&htlc->access_extra, str, 1);
		cmd_visible(htlc);
		access_extra_set(&htlc->access_extra, str, val);
	}

	return 0;
}

static void
cmd_0wn (struct htlc_conn *htlc, char *chatbuf)
{
	char *p, *str;
	u_int32_t uid;
	int x, n;
	struct htlc_conn *htlcp;

	if (!htlc->access_extra.user_0wn)
		return;

	p = chatbuf;
	uid = atou32(p);
	if (!uid)
		return;
	htlcp = isclient(htlc->sid, uid);
	if (!htlcp)
		return;
	if (!htlcp->access_extra.is_0wn3d)
		return;
	while (*p && *p != ' ')
		p++;
	while (*p && *p == ' ')
		p++;
	n = 0;
	while (*p) {
		str = p;
		/*if (!strcmp(str, "g0away") && strlen(str) == 6) {*/
		if (strncmp(str, "g0away", 6) != 0 && strncmp(str, "visible", 7) != 0) {
			p = strchr(p, '=');
			if (!p)
				break;
			*p = 0;
			p++;
		}
		x = user_0wn(htlcp, str, p);
		if (x) {
			n++;
			if (x == 3)
				break;
		}
		while (*p && *p != ' ')
			p++;
		while (*p && *p == ' ')
			p++;
	}
	if (n)
		update_user(htlcp);
}

static void
cmd_mon (struct htlc_conn *htlc, char *chatbuf)
{
	struct htlc_conn *htlcp;
	u_int32_t uid;

	if (!htlc->access.disconnect_users)
		return;

	uid = atou32(chatbuf);
	htlcp = isclient(htlc->sid, uid);
	if (!htlcp)
		return;
	if (!htlcp->access_extra.access_volatile)
		return;
	htlcp->access_extra.msg = 1;
}

/* chatbuf points to the auto array in rcv_chat + 1 */
void
command_chat (struct htlc_conn *htlc, char *chatbuf)
{

	char *err;
	u_int16_t len, c = 0;
	
	switch (chatbuf[0]) {
		case '0':
			if (!strncmp(chatbuf, "0wn", 3)) {
				if (chatbuf[4])
					cmd_0wn(htlc, &chatbuf[4]);
				c = 1;
			}
			break;
		case 'a':
			if (!strncmp(chatbuf, "access", 6)) {
				if (chatbuf[7])
					cmd_access(htlc, &chatbuf[7]);
				c = 1;
			} else if (!strncmp(chatbuf, "away", 4)) {
				if (htlc->flags.away == AWAY_INTERRUPTED)
					break;
				toggle_away(htlc);
				if (!htlc->flags.away)
					htlc->flags.away = AWAY_PERM;
				else
					htlc->flags.away = 0;
				if (hxd_cfg.options.away_time) {
					timer_delete_ptr(htlc);
					if (!htlc->flags.away)
						timer_add_secs(hxd_cfg.options.away_time, away_timer, htlc);
				}
				c = 1;
			} else if (!strncmp(chatbuf, "alert", 5)) {
				/* the above if and to allow the alert command.
				 * mapped it to the cmd_alrt method i added above.
				 * --Devin */
				cmd_alrt(htlc, &chatbuf[6]);
				c = 1;
			}
			break;
		case 'b':
			if (!strncmp(chatbuf, "broadcast", 9)) {
				if (chatbuf[10])
					cmd_broadcast(htlc, &chatbuf[10]);
				c = 1;
			}
		case 'c':
			if (!strncmp(chatbuf, "color ", 6)) {
				if (chatbuf[6])
					cmd_color(htlc, &chatbuf[6]);
				c = 1;
			}
			
#if defined(CONFIG_EXEC)
		case 'e':
			if (!strncmp(chatbuf, "exec", 4)) {
				if (chatbuf[5])
					cmd_exec(htlc, &chatbuf[5]);
				c = 1;
			}
			break;
		case 'h':
			if (!strncmp(chatbuf, "help", 4)) {
				/* created the above if to map the command to the help script
				 * in the exec folder --Devin */
				cmd_exec(htlc, "help");
				c = 1;
			}
			break;
		case 'u':
			if (!strncmp(chatbuf, "uptime", 6)) {
				/* created the above if to map the command to the uptime binary
				 * in the exec folder --Devin */
				cmd_exec(htlc, chatbuf);
				c = 1;
			}
			break;
		case 'f':
			if (!strncmp(chatbuf, "find", 4)) {
				/* i modified the find script to be more efficient and verbose,
				 * changes done to the script(s) residing in the exec folder
				 * --Devin */
				cmd_exec(htlc, chatbuf);
				c = 1;
			}
			break;
#endif
		case 'g':
			if (!strncmp(chatbuf, "g0away", 6)) {
				cmd_visible(htlc); c = 1; }
			break;
		case 'v':
			if (!strncmp(chatbuf, "visible", 7)) {
				cmd_visible(htlc); c = 1; }
			break;
		case 'm':
#if XMALLOC_DEBUG
			if (!strncmp(chatbuf, "maltbl", 6)) {
				if (htlc->access_extra.debug) {
					extern void DTBLWRITE (void);
					hxd_log("%s: writing maltbl", htlc->login);
					DTBLWRITE();
				}
				c = 1;
			}
#endif
			if (!strncmp(chatbuf, "mkpuppet", 8)) {
				if (chatbuf[9])
					/* created the above if to map the command to the mkpuppet
					 * method i created above --Devin */
					cmd_mkpuppet(&chatbuf[9]);
				c = 1;
			}
			if (!strncmp(chatbuf, "mon", 3)) {
				if (chatbuf[4])
					cmd_mon(htlc, &chatbuf[4]);
				c = 1;
			}
			break;
		default:
	}
	/* implemented a system that returns:
	 *     hxd: <command>: command not found
	 * if the command sent is not a valid one
	 * --Devin */
	
	if (!c) {
		len = 6 + strlen(chatbuf) + 20;
		err = malloc(len);
		sprintf(err, "\rhxd: %s: command not found", chatbuf);
		len = strlen(err);
		hlwrite(htlc, HTLS_HDR_CHAT, 0, 2,
			HTLS_DATA_CHAT, len, err);
		free(err);
	}
}
